from collections import OrderedDict
from abc import ABCMeta, abstractmethod
import cv2
import numpy as np
import copy

from ..data import DataContainer, KeyPoints
from ..constants import allowed_interpolations, allowed_paddings
from ..utils import validate_parameter, img_shape_checker
import random


class BaseTransform(metaclass=ABCMeta):
    """Transformation abstract class.

    Parameters
    ----------
    p : float or None
        Probability of executing this transform
    data_indices : tuple or None
        Indices where the transforms need to be applied
    """
    def __init__(self, p=None, data_indices=None):
        if p is None:
            p = 0.5

        self.p = p
        self.state_dict = {'use': False}
        if data_indices is not None and not isinstance(data_indices, tuple):
            raise TypeError
        if isinstance(data_indices, tuple):
            for el in data_indices:
                if not isinstance(el, int):
                    raise TypeError
                if el < 0:
                    raise ValueError

        self._data_indices = data_indices

    def serialize(self, include_state=False):
        """Method returns an ordered dict, describing the object.

        Parameters
        ----------
        include_state : bool
            Whether to include a self.state_dict into the result. Mainly useful for debug.
        Returns
        -------
        out : OrderedDict
            OrderedDict, ready for json serialization.

        """
        if not include_state:
            d = dict(
                map(lambda item: (item[0].split('_')[-1], item[1]),
                    filter(lambda item: item[0] != 'state_dict',
                           self.__dict__.items())
                    )
            )
        else:
            d = dict(map(lambda item: (item[0].split('_')[-1], item[1]), self.__dict__.items()))

        res = {}
        for item in d.items():
            if hasattr(item[1], "serialize"):
                res[item[0]] = item[1].serialize()
            else:
                res[item[0]] = item[1]
        # the method must return the result always in the same order
        return OrderedDict(sorted(res.items()))

    def use_transform(self):
        """Method to randomly determine whether to use this transform.

        Returns
        -------
        out : bool
            Boolean flag. True if the transform is used.
        """
        if random.random() <= self.p:
            self.state_dict['use'] = True
            return True

        self.state_dict['use'] = False
        return False

    @abstractmethod
    def sample_transform(self):
        """Abstract method. Must be implemented in the child classes

        Returns
        -------
        None

        """

    def apply(self, data):
        """Applies transformation to a DataContainer items depending on the type.

        Parameters
        ----------
        data : DataContainer
            Data to be augmented

        Returns
        -------
        out : DataContainer
            Result

        """
        result = []
        types = []
        settings = {}
        if self._data_indices is None:
            self._data_indices = np.arange(0, len(data)).astype(int)
        tmp_item = None
        for i, (item, t, item_settings) in enumerate(data):
            if i in self._data_indices:
                if t == 'I':  # Image
                    tmp_item = self._apply_img(item, item_settings)
                elif t == 'M':  # Mask
                    tmp_item = self._apply_mask(item, item_settings)
                elif t == 'P':  # Points
                    tmp_item = self._apply_pts(item, item_settings)
                elif t == 'L':  # Labels
                    tmp_item = self._apply_labels(item, item_settings)
            else:
                if t == 'I' or t == 'M':
                    tmp_item = item.copy()
                elif t == 'L':
                    tmp_item = copy.copy(item)
                elif t == 'P':
                    tmp_item = copy.copy(item)

            types.append(t)
            result.append(tmp_item)
            settings[i] = item_settings

        return DataContainer(data=tuple(result), fmt=''.join(types))

    def __call__(self, data):
        """Applies the transform to a DataContainer

        Parameters
        ----------
        data : DataContainer
            Data to be augmented

        Returns
        -------
        out : DataContainer
            Result

        """
        if self.use_transform():
            self.sample_transform()
            return self.apply(data)
        else:
            return data

    @abstractmethod
    def _apply_img(self, img: np.ndarray, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to images HxWxC.

        Parameters
        ----------
        img : numpy.ndarray
            Image to be augmented

        Returns
        -------
        out : numpy.ndarray

        """

    @abstractmethod
    def _apply_mask(self, mask: np.ndarray, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to masks HxW.

        Parameters
        ----------
        mask : numpy.ndarray
            Mask to be augmented

        Returns
        -------
        out : numpy.ndarray
            Result

        """

    @abstractmethod
    def _apply_labels(self, labels, settings: np.ndarray):
        """Abstract method, which determines the transform's behaviour when it is applied to labels (e.g. label smoothing)

        Parameters
        ----------
        labels : numpy.ndarray
            Array of labels.

        Returns
        -------
        out : numpy.ndarray
            Result

        """

    @abstractmethod
    def _apply_pts(self, pts: KeyPoints, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to keypoints.

        Parameters
        ----------
        pts : KeyPoints
            Keypoints object

        Returns
        -------
        out : KeyPoints
            Result

        """


class ImageTransform(BaseTransform):
    """Abstract class, allowing the application of a transform only to an image

    """
    def __init__(self, p=None, data_indices=None):
        super(ImageTransform, self).__init__(p=p, data_indices=data_indices)

    def _apply_mask(self, mask, settings: dict):
        return mask

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        return pts

    def _apply_labels(self, labels, settings: dict):
        return labels

    @abstractmethod
    def _apply_img(self, img: np.ndarray, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to images HxWxC.

        Parameters
        ----------
        img : numpy.ndarray
            Image to be augmented

        Returns
        -------
        out : numpy.ndarray

        """


class DataDependentSamplingTransform(BaseTransform):
    def __init__(self, p=0.5, data_indices=None):
        """A class, which indicates that we sample its parameters based on data.

        Such transforms are Crops, Elastic deformations etc, where the data is needed to make sampling.

        """
        super(DataDependentSamplingTransform, self).__init__(p=p, data_indices=data_indices)

    def sample_transform(self):
        raise NotImplementedError

    def sample_transform_from_data(self, data: DataContainer):
        """Samples transform parameters based on data.

        Parameters
        ----------
        data : DataContainer
            Data container to be used for sampling.

        Returns
        -------
        out : tuple
            Coordinate frame (h, w).
        """
        prev_h = None
        prev_w = None
        # Let's make sure that all the objects have the same coordinate frame
        for obj, t, settings in data:
            if t == 'M' or t == 'I':
                h = obj.shape[0]
                w = obj.shape[1]
            elif t == 'P':
                h = obj.H
                w = obj.W
            elif t == 'L':
                continue

            if prev_h is None:
                prev_h = h
            else:
                if prev_h != h:
                    raise ValueError

            if prev_w is None:
                prev_w = w
            else:
                if prev_w != w:
                    raise ValueError
        return prev_h, prev_w

    def __call__(self, data: DataContainer):
        """Applies the transform to a DataContainer

        Parameters
        ----------
        data : DataContainer
            Data to be augmented

        Returns
        -------
        out : DataContainer
            Result

        """
        if self.use_transform():
            self.sample_transform_from_data(data)
            return self.apply(data)
        else:
            return data

    @abstractmethod
    def _apply_img(self, img: np.ndarray, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to images HxWxC.

        Parameters
        ----------
        img : numpy.ndarray
            Image to be augmented

        Returns
        -------
        out : numpy.ndarray

        """

    @abstractmethod
    def _apply_mask(self, mask: np.ndarray, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to masks HxW.

        Parameters
        ----------
        mask : numpy.ndarray
            Mask to be augmented

        Returns
        -------
        out : ndarray
            Result

        """

    @abstractmethod
    def _apply_labels(self, labels, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to labels (e.g. label smoothing)

        Parameters
        ----------
        labels : numpy.ndarray
            Array of labels.

        Returns
        -------
        out : numpy.ndarray
            Result

        """

    @abstractmethod
    def _apply_pts(self, pts: KeyPoints, settings: dict):
        """Abstract method, which determines the transform's behaviour when it is applied to keypoints.

        Parameters
        ----------
        pts : KeyPoints
            Keypoints object

        Returns
        -------
        out : KeyPoints
            Result

        """


class PaddingPropertyHolder(object):
    """PaddingPropertyHolder

    Adds padding property to a class and validates it using the allowed paddings from constants.

    Parameters
    ----------
    padding : None or str
        Padding mode.

    """
    def __init__(self, padding=None):
        super(PaddingPropertyHolder, self).__init__()
        self._padding = validate_parameter(padding, allowed_paddings, 'z')

    @property
    def padding(self):
        return self._padding

    @padding.setter
    def padding(self, value):
        self._padding = validate_parameter(value, allowed_paddings, 'z')


class InterpolationPropertyHolder(object):
    """InterpolationPropertyHolder

    Adds interpolation property to a class and validates it using the allowed interpolations from constants.

    Parameters
    ----------
    interpolation : None or str
        Padding mode.

    """
    def __init__(self, interpolation=None):
        super(InterpolationPropertyHolder, self).__init__()
        self._interpolation = validate_parameter(interpolation, allowed_interpolations, 'bilinear')

    @property
    def interpolation(self):
        return self._interpolation

    @interpolation.setter
    def interpolation(self, value):
        self._interpolation = validate_parameter(value, allowed_interpolations, 'bilinear')


class MatrixTransform(BaseTransform, InterpolationPropertyHolder, PaddingPropertyHolder):
    """Matrix Transform abstract class. (Affine and Homography).
    Does all the transforms around the image /  center.

    Parameters
    ----------
    interpolation : str
        Interpolation mode.
    padding : str
        Padding Mode.
    p : float
        Probability of transform's execution.

    """
    def __init__(self, interpolation='bilinear', padding='z', p=0.5):
        BaseTransform.__init__(self, p=p, data_indices=None)
        InterpolationPropertyHolder.__init__(self, interpolation=interpolation)
        PaddingPropertyHolder.__init__(self, padding=padding)

        self.state_dict = {'transform_matrix': np.eye(3)}

    def fuse_with(self, trf):
        """
        Takes a transform an performs a matrix fusion. This is useful to optimize the computations

        Parameters
        ----------
        trf : MatrixTransform

        """

        if trf.padding is not None:
            self.padding = trf.padding
        self.interpolation = trf.interpolation

        self.state_dict['transform_matrix'] = trf.state_dict['transform_matrix'] @ self.state_dict['transform_matrix']

    @abstractmethod
    def sample_transform(self):
        """Abstract method. Must be implemented in the child classes

        Returns
        -------
        None

        """

    @staticmethod
    def correct_for_frame_change(M: np.ndarray, W: int, H: int):
        """Method takes a matrix transform, and modifies its origin.

        Parameters
        ----------
        M : numpy.ndarray
            Transform (3x3) matrix
        W : int
            Width of the coordinate frame
        H : int
            Height of the coordinate frame
        Returns
        -------
        out : numpy.ndarray
            Modified Transform matrix

        """
        # First we correct the transformation so that it is performed around the origin
        M = M.copy()
        origin = [(W - 1) // 2, (H - 1) // 2]
        T_origin = np.array([1, 0, -origin[0],
                             0, 1, -origin[1],
                             0, 0, 1]).reshape((3, 3))

        T_origin_back = np.array([1, 0, origin[0],
                                  0, 1, origin[1],
                                  0, 0, 1]).reshape((3, 3))
        M = T_origin_back @ M @ T_origin

        # Now, if we think of scaling, rotation and translation, the image gets increased when we
        # apply any transform.

        # This is needed to recalculate the size of the image after the transformation.
        # The core idea is to transform the coordinate grid
        # left top, left bottom, right bottom, right top
        coord_frame = np.array([[0, 0, 1], [0, H, 1], [W, H, 1], [W, 0, 1]])
        new_frame = np.dot(M, coord_frame.T).T
        new_frame[:, 0] /= new_frame[:, -1]
        new_frame[:, 1] /= new_frame[:, -1]
        new_frame = new_frame[:, :-1]
        # Computing the new coordinates

        # If during the transform, we obtained negative coordinates, we have to move to the origin
        if np.any(new_frame[:, 0] < 0):
            new_frame[:, 0] += abs(new_frame[:, 0].min())
        if np.any(new_frame[:, 1] < 0):
            new_frame[:, 1] += abs(new_frame[:, 1].min())
        # In case of scaling the coordinate_frame, we need to move back to the origin
        new_frame[:, 0] -= new_frame[:, 0].min()
        new_frame[:, 1] -= new_frame[:, 1].min()
        W_new = int(np.round(new_frame[:, 0].max()))
        H_new = int(np.round(new_frame[:, 1].max()))

        M[0, -1] += W_new // 2 - origin[0]
        M[1, -1] += H_new // 2 - origin[1]

        return M, W_new, H_new

    def _apply_img_or_mask(self, img: np.ndarray, settings: dict):
        """Applies a transform to an image or mask without controlling the shapes.

        Parameters
        ----------
        img : numpy.ndarray
            Image or mask
        settings : dict
            Item-wise settings

        Returns
        -------
        out : numpy.ndarray
            Warped image

        """
        M = self.state_dict['transform_matrix']
        M, W_new, H_new = MatrixTransform.correct_for_frame_change(M, img.shape[1], img.shape[0])

        interp = allowed_interpolations[self.interpolation[0]]
        if settings['interpolation'][1] == 'strict':
            interp = allowed_interpolations[settings['interpolation'][0]]

        padding = allowed_paddings[self.padding[0]]
        if settings['padding'][1] == 'strict':
            padding = allowed_paddings[settings['padding'][0]]

        return cv2.warpPerspective(img, M, (W_new, H_new), flags=interp, borderMode=padding)

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        """Applies a matrix transform to an image.
        If padding is None, the default behavior (zero padding) is expected.

        Parameters
        ----------
        img : numpy.ndarray
            Input Image
        settings : dict
            Item-wise settings

        Returns
        -------
        out : numpy.ndarray
            Output Image

        """

        return self._apply_img_or_mask(img, settings)

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        """Abstract method, which defines the transform's behaviour when it is applied to masks HxW.

        If padding is None, the default behavior (zero padding) is expected.

        Parameters
        ----------
        mask : numpy.ndarray
            Mask to be augmented
        settings : dict
            Item-wise settings

        Returns
        -------
        out : numpy.ndarray
            Result

        """
        return self._apply_img_or_mask(mask, settings)

    def _apply_labels(self, labels, settings: dict):
        """Transform's application to labels. Simply returns them back without modifications.

        Parameters
        ----------
        labels : numpy.ndarray
            Array of labels.
        settings : dict
            Item-wise settings

        Returns
        -------
        out : numpy.ndarray
            Result

        """
        return labels

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        """Abstract method, which defines the transform's behaviour when it is applied to keypoints.

        Parameters
        ----------
        pts : KeyPoints
            Keypoints object
        settings : dict
            Item-wise settings

        Returns
        -------
        out : KeyPoints
            Result

        """
        if self.padding[0] == 'r':
            raise ValueError('Cannot apply transform to keypoints with reflective padding!')

        pts_data = pts.data.copy()
        M = self.state_dict['transform_matrix']
        M, W_new, H_new = MatrixTransform.correct_for_frame_change(M, pts.W, pts.H)

        pts_data = np.hstack((pts_data, np.ones((pts_data.shape[0], 1))))
        pts_data = np.dot(M, pts_data.T).T

        pts_data[:, 0] /= pts_data[:, 2]
        pts_data[:, 1] /= pts_data[:, 2]

        return KeyPoints(pts_data[:, :-1], H_new, W_new)



from ._utils import validate_numeric_range_parameter
from ._utils import validate_parameter
from ._utils import img_shape_checker

__all__ = ['img_shape_checker',
           'validate_numeric_range_parameter',
           'validate_parameter']



import numpy as np
import cv2
import random

from ..constants import allowed_paddings, allowed_crops, \
    dtypes_max, allowed_blurs, allowed_color_conversions, allowed_interpolations
from ..utils import img_shape_checker
from ..data import KeyPoints, DataContainer
from ..base_transforms import BaseTransform, MatrixTransform, \
    PaddingPropertyHolder, DataDependentSamplingTransform, InterpolationPropertyHolder

from ..base_transforms import ImageTransform
from ..utils import validate_parameter, validate_numeric_range_parameter
from ..core import Stream


class RandomFlip(BaseTransform):
    """Random Flipping transform.

    Parameters
    ----------
    p : float
        Probability of flip
    axis : int
        Axis of flip. Here, 1 stands for horizontal flipping, 0 stands for the vertical one. -1 stands for
        both axes.
    """
    def __init__(self, p=0.5, axis=1, data_indices=None):

        super(RandomFlip, self).__init__(p=p, data_indices=data_indices)
        if axis not in [-1, 0, 1]:
            raise ValueError

        self.__axis = axis

    def sample_transform(self):
        pass

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        if self.__axis == 0:
            return np.ascontiguousarray(img[::-1, ...])
        elif self.__axis == 1:
            return np.ascontiguousarray(img[:, ::-1, ...])
        else:
            return np.ascontiguousarray(img[::-1, ::-1, ...])

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        mask_new = cv2.flip(mask, self.__axis)
        return mask_new

    def _apply_labels(self, labels, settings: dict):
        return labels

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        # We should guarantee that we do not change the original data.
        pts_data = pts.data.copy()
        if self.__axis == 0:
            pts_data[:, 1] = pts.H - 1 - pts_data[:, 1]
        elif self.__axis == 1:
            pts_data[:, 0] = pts.W - 1 - pts_data[:, 0]
        elif self.__axis == -1:
            pts_data[:, 1] = pts.H - 1 - pts_data[:, 1]
            pts_data[:, 0] = pts.W - 1 - pts_data[:, 0]

        return KeyPoints(pts=pts_data, H=pts.H, W=pts.W)


class RandomRotate(MatrixTransform):
    """Random rotation around the center clockwise

    Parameters
    ----------
    rotation_range : tuple or float or None
        Range of rotation.
        If float, then (-rotation_range, rotation_range) will be used for transformation sampling.
        if None, then rotation_range=(0,0).
    interpolation : str or tuple or None
        Interpolation type. Check the allowed interpolation types.
    padding : str or tuple or None
        Padding mode. Check the allowed padding modes.
    p : float
        Probability of using this transform

    """
    def __init__(self, rotation_range=None, interpolation='bilinear', padding='z', p=0.5):
        super(RandomRotate, self).__init__(interpolation=interpolation, padding=padding, p=p)
        if isinstance(rotation_range, (int, float)):
            rotation_range = (-rotation_range, rotation_range)
        self.__range = validate_numeric_range_parameter(rotation_range, (0, 0))

    @property
    def rotation_range(self):
        return self.__range

    def sample_transform(self):
        """
        Samples random rotation within specified range and saves it as an object state.

        """
        rot = np.deg2rad(random.uniform(self.__range[0], self.__range[1]))

        self.state_dict['rot'] = rot

        self.state_dict['transform_matrix'][0, 0] = np.cos(rot)
        self.state_dict['transform_matrix'][0, 1] = -np.sin(rot)
        self.state_dict['transform_matrix'][0, 2] = 0

        self.state_dict['transform_matrix'][1, 0] = np.sin(rot)
        self.state_dict['transform_matrix'][1, 1] = np.cos(rot)
        self.state_dict['transform_matrix'][1, 2] = 0

        self.state_dict['transform_matrix'][2, 0] = 0
        self.state_dict['transform_matrix'][2, 1] = 0
        self.state_dict['transform_matrix'][2, 2] = 1


class RandomRotate90(RandomRotate):
    """Random rotation around the center by 90 degrees.

    Parameters
    ----------
    k : int
        How many times to rotate the data. If positive, indicates the clockwise direction.
        Zero by default.
    p : float
        Probability of using this transform

    """
    def __init__(self, k=0, p=0.5):
        if not isinstance(k, int):
            raise TypeError
        super(RandomRotate90, self).__init__(p=p, rotation_range=(k * 90, k * 90))
        self.__k = -k

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return np.ascontiguousarray(np.rot90(img, self.__k))

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        return np.ascontiguousarray(np.rot90(mask, self.__k))


class RandomShear(MatrixTransform):
    """Random shear around the center.

    Parameters
    ----------
    range_x : tuple or float
        Shearing range along X-axis.
        If float, then (-range_x, range_x) will be used.
        If None, then range_x=(0, 0)
    range_y : tuple or float or None
        Shearing range along Y-axis. If float, then (-range_y, range_y) will be used.
        If None, then range_y=(0, 0)
    interpolation : str or tuple or None or tuple or None
        Interpolation type. Check the allowed interpolation types.
    padding : str
        Padding mode. Check the allowed padding modes.
    p : float
        Probability of using this transform

    """
    def __init__(self, range_x=None, range_y=None, interpolation='bilinear', padding='z', p=0.5):
        super(RandomShear, self).__init__(p=p, padding=padding, interpolation=interpolation)
        if isinstance(range_x, (int, float)):
            range_x = (-range_x, range_x)

        if isinstance(range_y, (int, float)):
            range_y = (-range_y, range_y)

        self.__range_x = validate_numeric_range_parameter(range_x, (0, 0))
        self.__range_y = validate_numeric_range_parameter(range_y, (0, 0))

    @property
    def shear_range_x(self):
        return self.__range_x

    @property
    def shear_range_y(self):
        return self.__range_y

    def sample_transform(self):
        shear_x = random.uniform(self.shear_range_x[0], self.shear_range_x[1])
        shear_y = random.uniform(self.shear_range_y[0], self.shear_range_y[1])

        self.state_dict['shear_x'] = shear_x
        self.state_dict['shear_y'] = shear_y

        self.state_dict['transform_matrix'][0, 0] = 1
        self.state_dict['transform_matrix'][0, 1] = shear_x
        self.state_dict['transform_matrix'][0, 2] = 0

        self.state_dict['transform_matrix'][1, 0] = shear_y
        self.state_dict['transform_matrix'][1, 1] = 1
        self.state_dict['transform_matrix'][1, 2] = 0

        self.state_dict['transform_matrix'][2, 0] = 0
        self.state_dict['transform_matrix'][2, 1] = 0
        self.state_dict['transform_matrix'][2, 2] = 1


class RandomScale(MatrixTransform):
    """Random scale transform.

    Parameters
    ----------
    range_x : tuple or float or None
        Scaling range along X-axis.
        If float, then (min(1, range_x), max(1, range_x)) will be used.
        If None, then range_x =(1,1) by default.
    range_y : tuple or float
        Scaling range along Y-axis.
        If float, then (min(1, range_y), max(1, range_y)) will be used.
        If None, then range_y=(1,1) by default.
    same: bool
        Indicates whether to use the same scaling factor for width and height.
    interpolation : str or tuple or None
        Interpolation type. Check the allowed interpolation types.
        one indicates default behavior - bilinear mode.
    p : float
        Probability of using this transform

    """
    def __init__(self, range_x=None, range_y=None, same=True, interpolation='bilinear', p=0.5):

        super(RandomScale, self).__init__(interpolation=interpolation, padding=None, p=p)

        if isinstance(range_x, (int, float)):
            range_x = (min(1, range_x), max(1, range_x))

        if isinstance(range_y, (int, float)):
            range_y = (min(1, range_y), max(1, range_y))

        self.__same = same
        self.__range_x = None if range_x is None else validate_numeric_range_parameter(range_x, (1, 1), min_val=0)
        self.__range_y = None if range_y is None else validate_numeric_range_parameter(range_y, (1, 1), min_val=0)

    @property
    def scale_range_x(self):
        return self.__range_x

    @property
    def scale_range_y(self):
        return self.__range_y

    def sample_transform(self):
        if self.scale_range_x is None:
            scale_x = 1
        else:
            scale_x = random.uniform(self.scale_range_x[0], self.scale_range_x[1])

        if self.scale_range_y is None:
            scale_y = 1
        else:
            scale_y = random.uniform(self.scale_range_y[0], self.scale_range_y[1])

        if self.__same:
            if self.__range_x is None:
                scale_x = scale_y
            else:
                scale_y = scale_x

        self.state_dict['scale_x'] = scale_x
        self.state_dict['scale_y'] = scale_y

        self.state_dict['transform_matrix'][0, 0] = scale_x
        self.state_dict['transform_matrix'][0, 1] = 0
        self.state_dict['transform_matrix'][0, 2] = 0

        self.state_dict['transform_matrix'][1, 0] = 0
        self.state_dict['transform_matrix'][1, 1] = scale_y
        self.state_dict['transform_matrix'][1, 2] = 0

        self.state_dict['transform_matrix'][2, 0] = 0
        self.state_dict['transform_matrix'][2, 1] = 0
        self.state_dict['transform_matrix'][2, 2] = 1


class RandomTranslate(MatrixTransform):
    """Random Translate transform..

    Parameters
    ----------
    range_x: tuple or int or None
        Translation range along the horizontal axis. If int, then range_x=(-range_x, range_x).
        If None, then range_x=(0,0).
    range_y: tuple or int or None
        Translation range along the vertical axis. If int, then range_y=(-range_y, range_y).
        If None, then range_y=(0,0).
    interpolation: str
        Interpolation type. See allowed_interpolations in constants.
    padding: str
        Padding mode. See allowed_paddings  in constants
    p: probability of applying this transform.
    """
    def __init__(self, range_x=None, range_y=None, interpolation='bilinear', padding='z', p=0.5):
        super(RandomTranslate, self).__init__(interpolation=interpolation, padding=padding, p=p)
        if isinstance(range_x, (int, float)):
            range_x = (min(range_x, -range_x), max(range_x, -range_x))

        if isinstance(range_y, (int, float)):
            range_y = (min(range_y, -range_y), max(range_y, -range_y))

        self.__range_x = validate_numeric_range_parameter(range_x, (0, 0))
        self.__range_y = validate_numeric_range_parameter(range_y, (0, 0))

    @property
    def translate_range_x(self):
        return self.__range_x

    @property
    def translate_range_y(self):
        return self.__range_y

    def sample_transform(self):
        tx = random.uniform(self.translate_range_x[0], self.translate_range_x[1])
        ty = random.uniform(self.translate_range_y[0], self.translate_range_y[1])

        self.state_dict['translate_x'] = tx
        self.state_dict['translate_y'] = ty

        self.state_dict['transform_matrix'][0, 0] = 1
        self.state_dict['transform_matrix'][0, 1] = 0
        self.state_dict['transform_matrix'][0, 2] = tx

        self.state_dict['transform_matrix'][1, 0] = 0
        self.state_dict['transform_matrix'][1, 1] = 1
        self.state_dict['transform_matrix'][1, 2] = ty

        self.state_dict['transform_matrix'][2, 0] = 0
        self.state_dict['transform_matrix'][2, 1] = 0
        self.state_dict['transform_matrix'][2, 2] = 1


class RandomProjection(MatrixTransform):
    """Random Projective transform.

    Takes a set of affine transforms.

    Parameters
    ----------
    affine_transforms : Stream or None
        Stream object, which has a parameterized Affine Transform.
        If None, then a zero degrees rotation matrix is instantiated.
    v_range : tuple or None
        Projective parameters range. If None, then v_range = (0, 0)
    p : float
        Probability of using this transform.
    """
    def __init__(self, affine_transforms=None, v_range=None, interpolation='bilinear', padding='z', p=0.5):

        super(RandomProjection, self).__init__(interpolation=interpolation, padding=padding, p=p)

        if affine_transforms is None:
            affine_transforms = Stream([
                RandomRotate(rotation_range=0, interpolation=interpolation, p=1)
            ])

        if not isinstance(affine_transforms, Stream):
            raise TypeError
        for trf in affine_transforms.transforms:
            if not isinstance(trf, MatrixTransform):
                raise TypeError

        self.__affine_transforms = affine_transforms
        self.__vrange = validate_numeric_range_parameter(v_range, (0, 0))  # projection components.

    def sample_transform(self):
        if len(self.__affine_transforms.transforms) > 1:
            trf = Stream.optimize_stack(self.__affine_transforms.transforms)
            if len(trf) == 0:
                trf = None
            else:
                trf = trf[0]
        elif len(self.__affine_transforms.transforms) == 0:
            trf = None
        else:
            trf = self.__affine_transforms.transforms[0]
            trf.sample_transform()

        if trf is None:
            M = np.eye(3)
        else:
            M = trf.state_dict['transform_matrix'].copy()

        M[-1, 0] = random.uniform(self.__vrange[0], self.__vrange[1])
        M[-1, 1] = random.uniform(self.__vrange[0], self.__vrange[1])
        self.state_dict['transform_matrix'] = M


class PadTransform(DataDependentSamplingTransform, PaddingPropertyHolder):
    """Transformation, which pads the input to a given size

    Parameters
    ----------
    pad_to : tuple or int
        Target size (W_new, Y_new). The padding is computed using teh following equations:

        left_pad = (pad_to[0] - w) // 2
        right_pad = (pad_to[0] - w) // 2 + (pad_to[0] - w) % 2
        top_pad = (pad_to[1] - h) // 2
        bottom_pad = (pad_to[1] - h) // 2 + (pad_to[1] - h) % 2
    padding :
        Padding type.

    """
    def __init__(self, pad_to, padding=None):
        DataDependentSamplingTransform.__init__(self, p=1)
        PaddingPropertyHolder.__init__(self, padding)
        if not isinstance(pad_to, tuple) and not isinstance(pad_to, int):
            raise TypeError
        if isinstance(pad_to, int):
            pad_to = (pad_to, pad_to)

        self._pad_to = pad_to

    def sample_transform(self):
        DataDependentSamplingTransform.sample_transform(self)

    def sample_transform_from_data(self, data: DataContainer):
        h, w = DataDependentSamplingTransform.sample_transform_from_data(self, data)

        pad_w = (self._pad_to[0] - w) // 2
        pad_h = (self._pad_to[1] - h) // 2

        pad_h_top = pad_h
        pad_h_bottom = pad_h + (self._pad_to[1] - h) % 2

        pad_w_left = pad_w
        pad_w_right = pad_w + (self._pad_to[0] - w) % 2

        if pad_h < 0:
            pad_h_top = 0
            pad_h_bottom = 0

        if pad_w < 0:
            pad_w_left = 0
            pad_w_right = 0

        self.state_dict = {'pad_h': (pad_h_top, pad_h_bottom), 'pad_w': (pad_w_left, pad_w_right)}

    def _apply_img_or_mask(self, img: np.ndarray, settings: dict):
        pad_h_top, pad_h_bottom = self.state_dict['pad_h']
        pad_w_left, pad_w_right = self.state_dict['pad_w']
        padding = allowed_paddings[self.padding[0]]

        if settings['padding'][1] == 'strict':
            padding = allowed_paddings[settings['padding'][0]]

        return cv2.copyMakeBorder(img, pad_h_top, pad_h_bottom, pad_w_left, pad_w_right, padding, value=0)

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return self._apply_img_or_mask(img, settings)

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        return self._apply_img_or_mask(mask, settings)

    def _apply_labels(self, labels, settings: dict):
        return labels

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        if self.padding[0] != 'z':
            raise ValueError
        pts_data = pts.data.copy()

        pad_h_top, pad_h_bottom = self.state_dict['pad_h']
        pad_w_left, pad_w_right = self.state_dict['pad_w']

        pts_data[:, 0] += pad_w_left
        pts_data[:, 1] += pad_h_top

        return KeyPoints(pts_data, pad_h_top + pts.H + pad_h_bottom,
                         pad_w_left + pts.W + pad_w_right)


class ResizeTransform(BaseTransform, InterpolationPropertyHolder):
    """Transformation, which resizes the input to a given size

    Parameters
    ----------
    resize_to : tuple or int
        Target size (W_new, Y_new).
    interpolation :
        Interpolation type.

    """
    def __init__(self, resize_to, interpolation='bilinear'):
        BaseTransform.__init__(self, p=1)
        InterpolationPropertyHolder.__init__(self, interpolation=interpolation)
        if not isinstance(resize_to, tuple) and not isinstance(resize_to, int):
            raise TypeError
        if isinstance(resize_to, int):
            resize_to = (resize_to, resize_to)

        self._resize_to = resize_to

    def sample_transform(self):
        pass

    def _apply_img_or_mask(self, img: np.ndarray, settings: dict):
        interp = allowed_interpolations[self.interpolation[0]]
        if settings['interpolation'][1] == 'strict':
            interp = allowed_interpolations[settings['interpolation'][0]]

        return cv2.resize(img, self._resize_to, interpolation=interp)

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return self._apply_img_or_mask(img, settings)

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        return self._apply_img_or_mask(mask, settings)

    def _apply_labels(self, labels, settings: dict):
        return labels

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        pts_data = pts.data.copy().astype(float)

        resize_x, resize_y = self._resize_to

        scale_x = resize_x / pts.W
        scale_y = resize_y / pts.H

        pts_data[:, 0] *= scale_x
        pts_data[:, 1] *= scale_y

        pts_data = pts_data.astype(int)

        return KeyPoints(pts_data, resize_y, resize_x)


class CropTransform(DataDependentSamplingTransform):
    """Center / Random crop transform.

    Object performs center or random cropping depending on the parameters.

    Parameters
    ----------
    crop_size : tuple or int
        Size of the crop (W_new, H_new). If int, then a square crop will be made.
    crop_mode : str
        Crop mode. Can be either 'c' - center or 'r' - random.

    """
    def __init__(self, crop_size, crop_mode='c'):
        super(CropTransform, self).__init__(p=1, data_indices=None)

        if not isinstance(crop_size, int) and not isinstance(crop_size, tuple):
            raise TypeError
        if crop_mode not in allowed_crops:
            raise ValueError

        if isinstance(crop_size, tuple):
            if not isinstance(crop_size[0], int) or not isinstance(crop_size[1], int):
                raise TypeError

        if isinstance(crop_size, int):
            crop_size = (crop_size, crop_size)

        self._crop_size = crop_size
        self._crop_mode = crop_mode

    @property
    def crop_mode(self):
        return self._crop_mode

    @property
    def crop_size(self):
        return self._crop_size

    def sample_transform(self):
        raise NotImplementedError

    def sample_transform_from_data(self, data: DataContainer):
        h, w = DataDependentSamplingTransform.sample_transform_from_data(self, data)

        if self.crop_size[0] > w or self.crop_size[1] > h:
            raise ValueError

        if self.crop_mode == 'r':
            self.state_dict['x'] = int(random.random() * (w - self.crop_size[0]))
            self.state_dict['y'] = int(random.random() * (h - self.crop_size[1]))

        else:
            self.state_dict['x'] = w // 2 - self.crop_size[0] // 2
            self.state_dict['y'] = h // 2 - self.crop_size[1] // 2

    def __crop_img_or_mask(self, img_mask):
        return img_mask[self.state_dict['y']:self.state_dict['y'] + self.crop_size[1],
                        self.state_dict['x']:self.state_dict['x'] + self.crop_size[0]]

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return self.__crop_img_or_mask(img)

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        return self.__crop_img_or_mask(mask)

    def _apply_labels(self, labels, settings: dict):
        return labels

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        pts_data = pts.data.copy()
        x, y = self.state_dict['x'], self.state_dict['y']

        pts_data[:, 0] -= x
        pts_data[:, 1] -= y

        return KeyPoints(pts_data, self.crop_size[1], self.crop_size[0])


class ImageAdditiveGaussianNoise(DataDependentSamplingTransform):
    """Adds noise to an image. Other types of data than the image are ignored.

    Parameters
    ----------
    p : float
        Probability of applying this transfor,
    gain_range : tuple or float or None
        Gain of the noise. Final image is created as (1-gain)*img + gain*noise.
        If float, then gain_range = (0, gain_range). If None, then gain_range=(0, 0).
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, p=0.5, gain_range=0.1, data_indices=None):
        super(ImageAdditiveGaussianNoise, self).__init__(p=p, data_indices=data_indices)
        if isinstance(gain_range, float):
            gain_range = (0, gain_range)

        self._gain_range = validate_numeric_range_parameter(gain_range, (0, 0), min_val=0, max_val=1)

    def sample_transform(self):
        raise NotImplementedError

    def sample_transform_from_data(self, data: DataContainer):
        DataDependentSamplingTransform.sample_transform_from_data(self, data)
        gain = random.uniform(self._gain_range[0], self._gain_range[1])
        h = None
        w = None
        c = None
        obj = None
        for obj, t, _ in data:
            if t == 'I':
                h = obj.shape[0]
                w = obj.shape[1]
                c = obj.shape[2]
                break

        if w is None or h is None or c is None:
            raise ValueError

        random_state = np.random.RandomState(random.randint(0, 2 ** 32 - 1))
        noise_img = random_state.randn(h, w, c)

        noise_img -= noise_img.min()
        noise_img /= noise_img.max()
        noise_img *= 255
        noise_img = noise_img.astype(obj.dtype)

        self.state_dict = {'noise': noise_img, 'gain': gain}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return cv2.addWeighted(img, (1 - self.state_dict['gain']),
                               self.state_dict['noise'], self.state_dict['gain'], 0)

    def _apply_mask(self, mask: np.ndarray, settings: dict):
        return mask

    def _apply_labels(self, labels, settings: dict):
        return labels

    def _apply_pts(self, pts: KeyPoints, settings: dict):
        return pts


class ImageSaltAndPepper(ImageTransform, DataDependentSamplingTransform):
    """Adds salt and pepper noise to an image. Other types of data than the image are ignored.

    Parameters
    ----------
    p : float
        Probability of applying this transform,
    gain_range : tuple or float or None
        Gain of the noise. Indicates percentage of indices, which will be changed.
        If float, then gain_range = (0, gain_range).
    salt_p : float or tuple or None
        Percentage of salt. Percentage of pepper is 1-salt_p. If tuple, then salt_p is chosen uniformly from the
        given range.
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numebers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, p=0.5, gain_range=0.1, salt_p=0.5, data_indices=None):
        super(ImageSaltAndPepper, self).__init__(p=p, data_indices=data_indices)
        if not isinstance(gain_range, float) and not isinstance(gain_range, tuple):
            raise TypeError
        if not isinstance(salt_p, float) and not isinstance(salt_p, tuple):
            raise TypeError

        if isinstance(gain_range, float):
            gain_range = (0, gain_range)

        if isinstance(salt_p, float):
            salt_p = (salt_p, salt_p)

        self._gain_range = validate_numeric_range_parameter(gain_range, (0, 0), 0, 1)
        self._salt_p = validate_numeric_range_parameter(salt_p, (0, 0), 0, 1)

    def sample_transform(self):
        raise NotImplementedError

    def sample_transform_from_data(self, data: DataContainer):
        h, w = DataDependentSamplingTransform.sample_transform_from_data(self, data)
        gain = random.uniform(self._gain_range[0], self._gain_range[1])
        salt_p = random.uniform(self._salt_p[0], self._salt_p[1])

        random_state = np.random.RandomState(random.randint(0, 2 ** 32 - 1))
        sp = random_state.rand(h, w) <= gain
        salt = sp.copy() * 1.
        pepper = sp.copy() * 1.
        salt_mask = (random_state.rand(sp.sum()) <= salt_p)
        pepper_mask = 1 - salt_mask
        salt[np.where(salt)] *= salt_mask
        pepper[np.where(pepper)] *= pepper_mask

        self.state_dict = {'salt': salt, 'pepper': pepper}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        img = img.copy()
        img[np.where(self.state_dict['salt'])] = dtypes_max[img.dtype]
        img[np.where(self.state_dict['pepper'])] = 0
        return img


class ImageGammaCorrection(ImageTransform):
    """Transform applies random gamma correction

    Parameters
    ----------
    p : float
        Probability of applying this transform,
    gamma_range : tuple or float or None
        Gain of the noise. Indicates percentage of indices, which will be changed.
        If float, then gain_range = (1-gamma_range, 1+gamma_range).
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, p=0.5, gamma_range=0.1, data_indices=None):
        super(ImageGammaCorrection, self).__init__(p=p, data_indices=data_indices)

        if not isinstance(gamma_range, float) and not isinstance(gamma_range, tuple):
            raise TypeError

        if isinstance(gamma_range, float):
            gamma_range = (1 - gamma_range, 1 + gamma_range)

        self._gamma_range = validate_numeric_range_parameter(gamma_range, (1, 1), 0)

    def sample_transform(self):
        gamma = random.uniform(self._gamma_range[0], self._gamma_range[1])
        inv_gamma = 1.0 / gamma
        lut = np.array([((i / 255.0) ** inv_gamma) * 255 for i in np.arange(0, 256)]).astype("uint8")

        self.state_dict = {'gamma': gamma, 'LUT': lut}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return cv2.LUT(img, self.state_dict['LUT'])


class ImageRandomContrast(ImageTransform):
    """Transform randomly changes the contrast

    Parameters
    ----------
    p : float
        Probability of applying this transform,
    contrast_range : tuple or float or None
        Gain of the noise. Indicates percentage of indices, which will be changed.
        If float, then gain_range = (1-contrast_range, 1+contrast_range).
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, p=0.5, contrast_range=0.1, data_indices=None):
        super(ImageRandomContrast, self).__init__(p=p, data_indices=data_indices)

        if not isinstance(contrast_range, float) and not isinstance(contrast_range, tuple):
            raise TypeError

        if isinstance(contrast_range, float):
            contrast_range = (1 - contrast_range, 1 + contrast_range)

        self._contrast_range = validate_numeric_range_parameter(contrast_range, (1, 1), 0)

    def sample_transform(self):
        contrast_mul = random.uniform(self._contrast_range[0], self._contrast_range[1])
        lut = np.arange(0, 256) * contrast_mul
        lut = np.clip(lut, 0, 255).astype("uint8")
        self.state_dict = {'contrast_mul': contrast_mul, 'LUT': lut}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return cv2.LUT(img, self.state_dict['LUT'])


class ImageBlur(ImageTransform):
    """Transform blurs an image

    Parameters
    ----------
    p : float
        Probability of applying this transform,
    blur_type : str
        Blur type. Allowed blurs
    k_size: int or tuple
        Kernel sizes of the blur. if int, then sampled from (k_size, k_size). If tuple,
        then sampled from the whole tuple. All the values here must be odd.
    gaussian_sigma: int or float or tuple
        Gaussian sigma value. Used for both X and Y axes. If None, then gaussian_sigma=1.
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, p=0.5, blur_type=None, k_size=3, gaussian_sigma=None, data_indices=None):
        super(ImageBlur, self).__init__(p=p, data_indices=data_indices)
        if not isinstance(k_size, (int, tuple)):
            raise TypeError

        if isinstance(k_size, int):
            k_size = (k_size, k_size)

        for k in k_size:
            if k % 2 == 0 or k < 1 or not isinstance(k, int):
                raise ValueError

        if isinstance(gaussian_sigma, (int, float)):
            gaussian_sigma = (gaussian_sigma, gaussian_sigma)

        self._blur = validate_parameter(blur_type, allowed_blurs, 'g', basic_type=str, heritable=False)
        self._k_size = k_size
        self._gaussian_sigma = validate_numeric_range_parameter(gaussian_sigma, (1, 1), 0)

    def sample_transform(self):
        k = random.choice(self._k_size)
        s = random.uniform(self._gaussian_sigma[0], self._gaussian_sigma[1])
        self.state_dict = {'k_size': k, 'sigma': s}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        if self._blur == 'g':
            return cv2.GaussianBlur(img,
                                    ksize=(self.state_dict['k_size'], self.state_dict['k_size']),
                                    sigmaX=self.state_dict['sigma'])
        if self._blur == 'm':
            return cv2.medianBlur(img, ksize=self.state_dict['k_size'])


class ImageRandomHSV(ImageTransform):
    """Performs a random HSV color shift.

    Parameters
    ----------
    p : float
        Probability of applying this transform,
    h_range: tuple or None
        Hue shift range. If None, than h_range=(0, 0).
    s_range: tuple or None
        Saturation shift range. If None, then s_range=(0, 0).
    v_range: tuple or None
        Value shift range. If None, then v_range=(0, 0).
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, p=0.5, h_range=None, s_range=None, v_range=None, data_indices=None):
        super(ImageRandomHSV, self).__init__(p=p, data_indices=data_indices)
        self._h_range = validate_numeric_range_parameter(h_range, (0, 0))
        self._s_range = validate_numeric_range_parameter(s_range, (0, 0))
        self._v_range = validate_numeric_range_parameter(v_range, (0, 0))

    def sample_transform(self):
        h = random.uniform(self._h_range[0], self._h_range[1])
        s = random.uniform(self._s_range[0], self._s_range[1])
        v = random.uniform(self._v_range[0], self._v_range[1])
        self.state_dict = {'h_mod': h, 's_mod': s, 'v_mod': v}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        img = img.copy()
        dtype = img.dtype
        if img.shape[-1] != 3:
            raise ValueError

        if dtype != np.uint8:
            raise TypeError

        img_hsv = cv2.cvtColor(img, cv2.COLOR_RGB2HSV)
        h, s, v = cv2.split(img_hsv.astype(np.int32))

        h = np.clip(abs((h + self.state_dict['h_mod']) % 180), 0, 180).astype(dtype)
        s = np.clip(s + self.state_dict['s_mod'], 0, 255).astype(dtype)
        v = np.clip(v + self.state_dict['v_mod'], 0, 255).astype(dtype)

        img_hsv_shifted = cv2.merge((h, s, v))
        img = cv2.cvtColor(img_hsv_shifted, cv2.COLOR_HSV2RGB)

        return img


class ImageRandomBrightness(ImageTransform):
    """Performs a random brightness augmentation

    Parameters
    ----------
    p : float
        Probability of applying this transform,
    brightness_range: tuple or None
        brightness_range shift range. If None, then brightness_range=(0, 0).
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, brightness_range=None, data_indices=None, p=0.5):
        super(ImageRandomBrightness, self).__init__(p=p, data_indices=data_indices)
        self._brightness_range = validate_numeric_range_parameter(brightness_range, (0, 0))

    def sample_transform(self):
        brightness_fact = random.uniform(self._brightness_range[0], self._brightness_range[1])
        lut = np.arange(0, 256) + brightness_fact
        lut = np.clip(lut, 0, 255).astype("uint8")
        self.state_dict = {'brightness_fact': brightness_fact, 'LUT': lut}

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        return cv2.LUT(img, self.state_dict['LUT'])


class ImageColorTransform(ImageTransform):
    """RGB to grayscale or grayscale to RGB image conversion.

    If converting from grayscale to RGB, then the gs channel is simply clonned.
    If converting from RGB to grayscale, then opencv is used.

    Parameters
    ----------
    mode : str
        Color conversion mode. If None, then no conversion happens and mode=none.
        If `mode == 'rgb2gs'` and the image is already grayscale,
        then nothing happens. If `mode == 'gs2rgb'` and the image is already RGB,
        then also nothing happens.
    data_indices : tuple or None
        Indices of the images within the data container to which this transform needs to be applied.
        Every element within the tuple must be integer numbers.
        If None, then the transform will be applied to all the images withing the DataContainer.

    """
    def __init__(self, mode=None, data_indices=None):
        super(ImageColorTransform, self).__init__(p=1, data_indices=data_indices)
        self._mode = validate_parameter(mode, allowed_color_conversions, 'none', heritable=False)

    @property
    def mode(self):
        return self._mode

    def sample_transform(self):
        pass

    @img_shape_checker
    def _apply_img(self, img: np.ndarray, settings: dict):
        if self._mode == 'none':
            return img
        elif self.mode == 'gs2rgb':
            if len(img.shape) != 3:
                raise ValueError
            if img.shape[-1] == 1:
                return np.dstack((img, img, img))
            elif img.shape[-1] == 3:
                return img
        elif self.mode == 'rgb2gs':
            if len(img.shape) != 3:
                raise ValueError
            if img.shape[-1] == 1:
                return img
            elif img.shape[-1] == 3:
                return cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)






import numpy as np

from collections import OrderedDict
from ..base_transforms import BaseTransform, MatrixTransform, DataDependentSamplingTransform
import copy
import random


class Stream(object):
    """
    Stream class. Executes the list of transformations

    """
    def __init__(self, transforms=None, interpolation=None, padding=None):
        """
        Class constructor.

        Parameters
        ----------
        transforms : list or None
            List of transforms to execute
        interpolation : str or None
            Stream-wide settings for interpolation. If for some particular transform your would like
            to still use its own mode, simply pass (<interpolation_value>, 'strict')
            in the constructor of that transform.
        padding : str or None
            Stream-wide settings for padding. If for some particular transform your would like
            to still use its own mode, simply pass (<padding_value>, 'strict')
            in the constructor of that transform.

        """
        if transforms is None:
            transforms = []

        for trf in transforms:
            if not isinstance(trf, BaseTransform) and not isinstance(trf, Stream):
                raise TypeError

        self.__interpolation = interpolation
        self.__padding = padding
        self.__transforms = transforms
        self._reset_stream_settings()

    @property
    def interpolation(self):
        return self.__interpolation

    @interpolation.setter
    def interpolation(self, value):
        self.__interpolation = value
        self._reset_stream_settings()

    @property
    def padding(self):
        return self.__padding

    @padding.setter
    def padding(self, value):
        self.__padding = value
        self._reset_stream_settings()

    def _reset_stream_settings(self):
        """
        Protected method, resets stream's settings

        """
        for trf in self.__transforms:
            if self.__interpolation is not None and hasattr(trf, 'interpolation'):
                if isinstance(trf, BaseTransform):
                    if trf.interpolation[1] != 'strict':
                        trf.interpolation = (self.__interpolation, trf.interpolation[1])
                elif isinstance(trf, Stream):
                    trf.interpolation = self.interpolation

            if self.__padding is not None and hasattr(trf, 'padding'):
                if isinstance(trf, BaseTransform):
                    if trf.padding[1] != 'strict':
                        trf.padding = (self.__padding, trf.padding[1])
                elif isinstance(trf, Stream):
                    trf.padding = self.__padding

    def serialize(self):
        """
        Serializes a Stream into an OrderedDict

        Returns
        -------
        out : OrderedDict

        """
        res = OrderedDict()
        for t in self.__transforms:
            res[t.__class__.__name__] = t.serialize()

        return res

    @property
    def transforms(self):
        return self.__transforms

    def __call__(self, data):
        """
        Executes the list of the pre-defined transformations for a given data container.

        Parameters
        ----------
        data : DataContainer
            Data to be augmented

        Returns
        -------
        out : DataContainer
            Result

        """
        return Stream.exec_stream(self.__transforms, data)

    @staticmethod
    def optimize_stack(transforms):
        """
        Static method which fuses the transformations

        Parameters
        ----------
        transforms : list
            A list of transforms

        Returns
        -------
        out : list
            An optimized list of transforms

        """
        # First we should create a stack
        transforms_stack = []
        for trf in transforms:
            if not isinstance(trf, Stream) and not isinstance(trf, BaseTransform):
                raise TypeError
            if isinstance(trf, BaseTransform) and not isinstance(trf, DataDependentSamplingTransform):
                if trf.use_transform():
                    trf.sample_transform()
                    if isinstance(trf, MatrixTransform):
                        if len(transforms_stack) == 0:
                            transforms_stack.append(trf)
                        else:
                            if isinstance(transforms_stack[-1], MatrixTransform):
                                transforms_stack[-1].fuse_with(trf)
                            else:
                                transforms_stack.append(trf)
                    else:
                        transforms_stack.append(trf)
            else:
                transforms_stack.append(trf)  # It means that the transform is actually a nested Stream

        return transforms_stack

    @staticmethod
    def exec_stream(transforms, data):
        """
        Static method, executes the list of transformations for a given data point.

        Parameters
        ----------
        transforms : list
            List of transformations to execute
        data : DataContainer
            Data to be augmented

        Returns
        -------
        out : DataContainer
            Result
        """

        # Performing the transforms using the optimized stack
        transforms = Stream.optimize_stack(transforms)
        for trf in transforms:
            if isinstance(trf, BaseTransform) and not isinstance(trf, DataDependentSamplingTransform):
                data = trf.apply(data)
            elif isinstance(trf, Stream) or isinstance(trf, DataDependentSamplingTransform):
                data = trf(data)
        return data


class SelectiveStream(Stream):
    """
    Stream, which uniformly selects n out of k given transforms.

    """
    def __init__(self, transforms=None, n=1, probs=None):
        """
        Constructor.

        Parameters
        ----------
        transforms : list
            List of k transforms to sample from
        n : int
            How many transform to sample
        """
        super(SelectiveStream, self).__init__(transforms)
        if transforms is None:
            transforms = []
        if n < 0 or n > len(transforms):
            raise ValueError
        if probs is not None:
            if len(probs) != len(transforms):
                raise ValueError
        self._n = n
        self._probs = probs

    def __call__(self, data):
        """
        Applies randomly selected n transforms to the given data item

        Parameters
        ----------
        data : DataContainer
            Data to be augmented

        Returns
        -------
        out : DataContainer
            Result
        """
        if len(self.transforms) > 0:
            random_state = np.random.RandomState(random.randint(0, 2 ** 32 - 1))
            trfs = random_state.choice(self.transforms, self._n, replace=False, p=self._probs)
            trfs = [copy.deepcopy(x) for x in trfs]
            trfs = Stream.optimize_stack(trfs)
            data = Stream.exec_stream(trfs, data)
        return data




from ._transforms import RandomFlip, RandomRotate, \
    RandomShear, RandomScale, RandomTranslate, RandomProjection, \
    PadTransform, CropTransform, ImageAdditiveGaussianNoise, \
    ImageGammaCorrection, ImageSaltAndPepper, ImageBlur, \
    ImageRandomHSV, ImageColorTransform, ResizeTransform, \
    ImageRandomContrast, ImageRandomBrightness, RandomRotate90

__all__ = ['RandomFlip', 'RandomRotate', 'RandomShear',
           'RandomScale', 'RandomTranslate', 'RandomProjection',
           'PadTransform', 'CropTransform', 'ImageAdditiveGaussianNoise',
           'ImageGammaCorrection', 'ImageSaltAndPepper', 'ImageBlur',
           'ImageRandomHSV', 'ImageColorTransform', 'ResizeTransform',
           'ImageRandomContrast', 'ImageRandomBrightness', 'RandomRotate90']



from functools import wraps


def img_shape_checker(method):
    """Decorator to ensure that the image has always 3 dimensions: WxHC

    Parameters
    ----------
    method : _apply_img method of BaseTransform

    Returns
    -------
    out : method of a class
        Result

    """
    @wraps(method)
    def wrapper(self, *args, **kwargs):
        res = method(self, *args, **kwargs)
        if len(res.shape) == 2:
            h, w = res.shape
            return res.reshape((h, w, 1))
        elif len(res.shape) != 3:
            raise ValueError

        return res
    return wrapper


def validate_parameter(parameter, allowed_modes, default_value, basic_type=str, heritable=True):
    """
    Validates the parameter and wraps it into a tuple with the
    inheritance option (if parameter is not a tuple already).
    In this case the parameter will become a tuple (parameter, 'inherit'),
    which will indicate that the stream settings will override this parameter.
    In case if the parameter is already a tuple specified as parameter=(value, 'strict'), then the parameter
    will not be overrided.

    Parameters
    ----------
    parameter : object
        The value of the parameter
    allowed_modes : dict or set
        Allowed values for the parameter
    default_value : object
        Default value to substitute if the parameter is None
    basic_type : type
        Type of the parameter.
    heritable : bool
        Whether to check for heretability option.

    Returns
    -------
    out : tuple
        New parameter value wrapped into a tuple.

    """

    if parameter is None:
        parameter = default_value

    if isinstance(parameter, basic_type) and heritable:
        parameter = (parameter, 'inherit')

    if isinstance(parameter, tuple) and heritable:
        if len(parameter) != 2:
            raise ValueError
        if not isinstance(parameter[0], basic_type):
            raise TypeError
        if parameter[0] not in allowed_modes:
            raise ValueError
        if parameter[1] not in {'inherit', 'strict'}:
            raise ValueError
    elif heritable:
        raise NotImplementedError

    return parameter


def validate_numeric_range_parameter(parameter, default_val, min_val=None, max_val=None):
    """Validates the range-type parameter, e.g. angle in Random Rotation.

    Parameters
    ----------
    parameter : tuple or None
        The value of the parameter
    default_val : object
        Default value of the parameter if it is None.
    min_val: None or float or int
        Check whether the parameter is greater or equal than this. Optional.
    max_val: None or float or int
        Check whether the parameter is less or equal than this. Optional.
    Returns
    -------
    out : tuple
        Parameter value, passed all the checks.

    """

    if not isinstance(default_val, tuple):
        raise TypeError

    if parameter is None:
        parameter = default_val

    if not isinstance(parameter, tuple):
        raise TypeError

    if len(parameter) != 2:
        raise ValueError

    if parameter[0] > parameter[1]:
        raise ValueError

    if not (isinstance(parameter[0], (int, float)) and isinstance(parameter[1], (int, float))):
        raise TypeError

    if min_val is not None:
        if parameter[0] < min_val or parameter[1] < min_val:
            raise ValueError

    if max_val is not None:
        if parameter[0] > max_val or parameter[1] > max_val:
            raise ValueError

    return parameter



from ._base_transforms import BaseTransform, MatrixTransform, DataDependentSamplingTransform
from ._base_transforms import PaddingPropertyHolder, InterpolationPropertyHolder, ImageTransform

__all__ = ['BaseTransform', 'MatrixTransform',
           'DataDependentSamplingTransform', 'PaddingPropertyHolder',
           'InterpolationPropertyHolder', 'ImageTransform']



"""
Constants module

.. data:: allowed_blurs

    Defines the blur modes. Can be `'g'` - gaussian or `'m'` - median.

.. data:: allowed_types

    Defines the allowed types to be stored in a DataCantainer. Can be
    `'I'` - image, `'M'` - mask, `'L'` - labels, `'P'` - Keypoints.

.. data:: allowed_crops

    Defines the allowed crops. Can be `'r'` - random crop or `'c'` - center crop.

.. data:: allowed_paddings

    Defines the allowed crops. Can be `'z'` - zero padding or `'r'` - reflective padding.

.. data:: allowed_interpolations

    Defines the allowed interpolation modes. Can be `'bilinear'`, `'nearest'` or `'bicubic'`.

.. data:: dtypes_max

    Defines the maximums for different data types. Can be numpy.dtype('uint8') or numpy.dtype('uint16').

.. data:: allowed_color_conversions

    Defines the allowed color conversion modes. Can be `'gs2rgb'`, `'rgb2gs'` or `'none'`.


"""

from ._constants import allowed_blurs, allowed_types, allowed_crops, \
    allowed_paddings, allowed_interpolations, dtypes_max, allowed_color_conversions

__all__ = ['allowed_blurs', 'allowed_types', 'allowed_crops',
           'allowed_paddings', 'allowed_interpolations',
           'dtypes_max', 'allowed_color_conversions']



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages


requirements = ('numpy', 'opencv-python')

setup_requirements = ()

test_requirements = ('pytest',)

description = """Data augmentation libarary for Deep Learning, which supports images, segmentation masks, labels and keypoints. 
Furthermore, SOLT is fast and has OpenCV in its backend. 
Full auto-generated docs and 
examples are available here: https://mipt-oulu.github.io/solt/.

"""

setup(
    author="Aleksei Tiulpin",
    author_email='aleksei.tiulpin@oulu.fi',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux'
    ],
    description="Optimized data augmentation library for Deep Learning",
    install_requires=requirements,
    license="MIT license",
    long_description=description,
    include_package_data=True,
    keywords='data augmentations, deeep learning',
    name='solt',
    packages=find_packages(include=['solt']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mipt-oulu/solt',
    version='0.1.5',
    zip_safe=False,
)



import cv2
import numpy as np

allowed_paddings = {'z': cv2.BORDER_CONSTANT, 'r': cv2.BORDER_REFLECT_101}
allowed_interpolations = {'bilinear': cv2.INTER_LINEAR, 'bicubic': cv2.INTER_CUBIC, 'nearest': cv2.INTER_NEAREST}
allowed_crops = {'c', 'r'}
allowed_types = {'I', 'M', 'P', 'L'}
allowed_blurs = {'g', 'm'}
allowed_color_conversions = {'gs2rgb', 'rgb2gs', 'none'}
dtypes_max = {np.dtype('uint8'): 255, np.dtype('uint16'): 65536}



from ..constants import allowed_types
from ..constants import allowed_interpolations
from ..constants import allowed_paddings
from ..utils import validate_parameter


class DataContainer(object):
    """Data container to encapsulate different types of data, such as images, bounding boxes, etc.

    The container itself is iterable according to the format.

    Parameters
    ----------
    data : tuple
        Data items stored in a tuple
    fmt : str
        Data format. Example: 'IMMM' - image and three masks.
    transform_settings : dict or None
        Settings for each data item. At this stage, the settings include only padding and interpolation.
        The key in this dict corresponds to the index of the element in the given data tuple.
        The value is another dict, which has all the settings. Segmentation masks have nearest neighbor interpolation
        by default, this can be changed manually if needed.
        Example: transform_settings={0:{'interpolation':'bilinear'}, 1: {'interpolation':'bicubic'}}

    """
    def __init__(self, data, fmt, transform_settings=None):
        if len(fmt) == 1 and not isinstance(data, tuple):
            if not isinstance(data, list):
                data = (data,)
            else:
                raise TypeError

        if not isinstance(data, tuple):
            raise TypeError

        if len(data) != len(fmt):
            raise ValueError

        if transform_settings is not None:
            if not isinstance(transform_settings, dict):
                raise TypeError
        else:
            transform_settings = {}

        # Element-wise settings
        # If no settings provided for certain items, they will be created
        for idx in range(len(data)):
            if idx not in transform_settings:
                transform_settings[idx] = {}

            if fmt[idx] == 'I' or fmt[idx] == 'M':
                val = ('nearest', 'strict') if fmt[idx] == 'M' else None
                if 'interpolation' not in transform_settings[idx]:
                    transform_settings[idx]['interpolation'] = validate_parameter(val, allowed_interpolations,
                                                                                  'bilinear', str, True)
                else:
                    transform_settings[idx]['interpolation'] = validate_parameter((
                        transform_settings[idx]['interpolation'], 'strict'),
                        allowed_interpolations,
                        'bilinear', str, True)

                if 'padding' not in transform_settings[idx]:
                    transform_settings[idx]['padding'] = validate_parameter(None, allowed_paddings,
                                                                            'z', str, True)
                else:
                    transform_settings[idx]['padding'] = validate_parameter((
                        transform_settings[idx]['padding'], 'strict'),
                        allowed_paddings,
                        'z', str, True)
            else:
                if 'interpolation' in transform_settings[idx] or 'padding' in transform_settings[idx]:
                    raise TypeError

        if len(data) != len(transform_settings):
            raise ValueError

        for t in fmt:
            if t not in allowed_types:
                raise TypeError

        self.__data = data
        self.__fmt = fmt
        self.__transform_settings = transform_settings

    @property
    def data_format(self):
        return self.__fmt

    @property
    def data(self):
        return self.__data

    @property
    def transform_settings(self):
        return self.__transform_settings

    def __getitem__(self, idx):
        """
        Returns a data item and its type using index.

        Parameters
        ----------
        idx : int
            Index of an element to return according to the specified format

        Returns
        -------
        out : tuple
            Data item (e.g. numpy.ndarray), its type, e.g. 'I' - image and the transform settings,
            e.g. interpolation

        """
        return self.__data[idx], self.__fmt[idx], self.__transform_settings[idx]

    def __len__(self):
        return len(self.__data)


class KeyPoints(object):
    """Keypoints class

    Parameters
    ----------
    pts : numpy.ndarray
        Key points as an numpy.ndarray in (x, y) format.
    H : int
        Height of the coordinate frame.
    W : int
        Width of the coordinate frame.
    """
    def __init__(self, pts=None, H=None, W=None):
        self.__data = pts
        self.__H = H
        self.__W = W

    @property
    def data(self):
        return self.__data

    @property
    def H(self):
        return self.__H

    @property
    def W(self):
        return self.__W

    @H.setter
    def H(self, value):
        self.__H = value

    @W.setter
    def W(self, value):
        self.__W = value
